#!/usr/local/bin/python
# -*- coding: utf-8 -*-
'''
@date: 09/2018
@author: cfraisse
'''

"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
This script defines fourteen models of population divergence used to perform joint demographic inferences with ∂a∂i v1.6.3 (Gutenkunst, et al. 2009). 

There are seven categories of models: "SI": Strict Isolation, "IM": Isolation with Migration, "AM": Ancient Migration, "SC": Secondary Contact, "PAM": Periodic Ancient Migration, "PSC": Periodic Secondary Contact, "AMSISC": three-period model.

Each model comes in subcategories: "no suffix": Reference model; "2N": Hill-Robertson effects with two categories of population size in the genome (we assume population size is reduced by a given factor in non-recombining regions); "2M": Selection against migrants with two categories of migration rate in the genome (we assume migration is null in barrier regions).

All models assume an initial ancestral population which is split in two daughter populations.
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

##################
# LIBRARY IMPORT #
##################
import sys
import dadi
import numpy


##########
# MODELS #
##########
"""""""""""""""""
Strict Isolation
"""""""""""""""""
def SI(params, (n1,n2), pts):
    nu1, nu2, Ts = params
    """
    Model with split and strict isolation.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    Ts: Time of divergence in strict isolation.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    # Define the grid we'll use
    xx = dadi.Numerics.default_grid(pts)
    # Ancestral population
    phi = dadi.PhiManip.phi_1D(xx)
    # Split event
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    # Divergence in strict isolation
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=0, m21=0)
    # Final spectrum
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def SI2N(params, (n1,n2), pts):
    nu1, nu2, Ts, nr, bf = params
    """
    Model with split and strict isolation; two categories of population size in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    Ts: Time of divergence in strict isolation.
    nr: Proportion of "non-recombining" regions affected by background selection.
    bf : Background factor, which defines the extent of population size reduction in "nr" regions. 
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-recombining regions
    phinr = dadi.PhiManip.phi_1D(xx)
    phinr = dadi.PhiManip.phi_1D_to_2D(xx, phinr)
    phinr = dadi.Integration.two_pops(phinr, xx, Ts, nu1*bf, nu2*bf, m12=0, m21=0)
    fsnr = dadi.Spectrum.from_phi(phinr, (n1,n2), (xx,xx))
    # Spectrum for recombining regions
    phir = dadi.PhiManip.phi_1D(xx)
    phir = dadi.PhiManip.phi_1D_to_2D(xx, phir)
    phir = dadi.Integration.two_pops(phir, xx, Ts, nu1, nu2, m12=0, m21=0)
    fsr = dadi.Spectrum.from_phi(phir, (n1,n2), (xx,xx))
    # Final spectrum
    fs= nr*fsnr + (1-nr)*fsr
    return fs


"""""""""""""""""""""""""""
Isolation with Migration
"""""""""""""""""""""""""""
def IM(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts = params
    """
    Model with continuous migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1.
    m21: Migration from population 1 to population 2.
    Ts: Time of divergence in continuous migration.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def IM2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, P1, P2 = params
    """
    Model with continuous migration; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Ts: Time of divergence in continuous migration.
    Te: Time of the exponential growth in continuous migration.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=m12, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=m21)
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs


"""""""""""""""""
Ancient Migration
"""""""""""""""""
def AM(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tam = params
    """
    Model with split, ancient migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1.
    m21: Migration from population 1 to population 2.
    Tam: Time of ancient migration.
    Ts: Time of divergence in strict isolation.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phi = dadi.Integration.two_pops(phi, xx, Tam, nu1, nu2, m12=0, m21=0)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def PAM(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tam = params
    """
    Model with split, followed by two periods of ancient migration.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1.
    m21: Migration from population 1 to population 2.
    Tam: Time of ancient migration.
    Ts: Time of divergence in strict isolation.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phi = dadi.Integration.two_pops(phi, xx, Tam, nu1, nu2, m12=0, m21=0)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phi = dadi.Integration.two_pops(phi, xx, Tam, nu1, nu2, m12=0, m21=0)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def AM2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tam, P1, P2 = params
    """
    Model with split, ancient migration; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Tam: Time of ancient migration.
    Ts: Time of divergence in strict isolation.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=m12, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=m21)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs


def PAM2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tam, P1, P2 = params
    """
    Model with split, followed by two preriods of ancient migration; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Tam: Time of ancient migration.
     Ts: Time of divergence in strict isolation.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=m12, m21=m21)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=m12, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=m12, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=m21)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=m21)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tam, nu1, nu2, m12=0, m21=0)
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs



"""""""""""""""""
Secondary Contact
"""""""""""""""""
def SC(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc = params
    """
    Model with split, strict isolation, and secondary contact.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1.
    m21: Migration from population 1 to population 2.
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=0, m21=0)
    phi = dadi.Integration.two_pops(phi, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs

    
def PSC(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc = params
    """
    Model with split, strict isolation, followed by two periods of secondary contact.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1.
    m21: Migration from population 1 to population 2.
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=0, m21=0)
    phi = dadi.Integration.two_pops(phi, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=0, m21=0)
    phi = dadi.Integration.two_pops(phi, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def SC2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc, P1, P2 = params
    """
    Model with split, strict isolation, and secondary contact; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tsc, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tsc, nu1, nu2, m12=m12, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tsc, nu1, nu2, m12=0, m21=m21)
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs


def PSC2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc, P1, P2 = params
    """
    Model with split, strict isolation, followed by two periods of secondary contact; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tsc, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tsc, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tsc, nu1, nu2, m12=m12, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tsc, nu1, nu2, m12=m12, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tsc, nu1, nu2, m12=0, m21=m21)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tsc, nu1, nu2, m12=0, m21=m21) 
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs



"""""""""""""""""""""
Three-period model
"""""""""""""""""""""
def AMSISC(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc, Tam = params
    """
    Model with split, followed by a period of ancient migration, complete isolation, and secondary contact.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Tam: Time of ancient migration.
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    n1,n2: Size of fs to generate. 
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    phi = dadi.PhiManip.phi_1D(xx)
    phi = dadi.PhiManip.phi_1D_to_2D(xx, phi)
    phi = dadi.Integration.two_pops(phi, xx, Tam, nu1, nu2, m12=m12, m21=m21)
    phi = dadi.Integration.two_pops(phi, xx, Ts, nu1, nu2, m12=0, m21=0)
    phi = dadi.Integration.two_pops(phi, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fs = dadi.Spectrum.from_phi(phi, (n1,n2), (xx,xx))
    return fs


def AMSISC2M(params, (n1,n2), pts):
    nu1, nu2, m12, m21, Ts, Tsc, Tam, P1, P2 = params
    """
    Model with split, followed by a period of ancient migration, complete isolation, and secondary contact; two categories of migration rate in the genome.

    nu1: Size of population 1 after split.
    nu2: Size of population 2 after split.
    m12: Migration from population 2 to population 1 in non-barrier regions (m=0 in barrier regions).
    m21: Migration from population 1 to population 2 in non-barrier regions (m=0 in barrier regions).
    Tam: Time of ancient migration.
    Ts: Time of divergence in strict isolation.
    Tsc: Time of secondary contact.
    P1: Proportion of "non-barrier" regions in population 1.
    P2: Proportion of "non-barrier" regions in population 2.
    n1,n2: Size of fs to generate.
    pts: Number of points to use in grid for evaluation.
    """
    xx = dadi.Numerics.default_grid(pts)
    # Spectrum for non-barrier regions in population 1 and 2
    phiN1N2 = dadi.PhiManip.phi_1D(xx)
    phiN1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1N2)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tam, nu1, nu2, m12=m12, m21=m21)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1N2 = dadi.Integration.two_pops(phiN1N2, xx, Tsc, nu1, nu2, m12=m12, m21=m21)
    fsN1N2 = dadi.Spectrum.from_phi(phiN1N2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and 2
    phiI1I2 = dadi.PhiManip.phi_1D(xx)
    phiI1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1I2)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tam, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1I2 = dadi.Integration.two_pops(phiI1I2, xx, Tsc, nu1, nu2, m12=0, m21=0)
    fsI1I2 = dadi.Spectrum.from_phi(phiI1I2, (n1,n2), (xx,xx))
    # Spectrum for non-barrier regions in population 1 and barrier regions in population 2
    phiN1I2 = dadi.PhiManip.phi_1D(xx)
    phiN1I2 = dadi.PhiManip.phi_1D_to_2D(xx, phiN1I2)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tam, nu1, nu2, m12=m12, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiN1I2 = dadi.Integration.two_pops(phiN1I2, xx, Tsc, nu1, nu2, m12=m12, m21=0)
    fsN1I2 = dadi.Spectrum.from_phi(phiN1I2, (n1,n2), (xx,xx))
    # Spectrum for barrier regions in population 1 and non-barrier regions in population 2
    phiI1N2 = dadi.PhiManip.phi_1D(xx)
    phiI1N2 = dadi.PhiManip.phi_1D_to_2D(xx, phiI1N2)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tam, nu1, nu2, m12=0, m21=m21)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Ts, nu1, nu2, m12=0, m21=0)
    phiI1N2 = dadi.Integration.two_pops(phiI1N2, xx, Tsc, nu1, nu2, m12=0, m21=m21) 
    fsI1N2 = dadi.Spectrum.from_phi(phiI1N2, (n1,n2), (xx,xx))
    # Final spectrum
    fs = P1*P2*fsN1N2 + (1-P1)*(1-P2)*fsI1I2 + P1*(1-P2)*fsN1I2 + (1-P1)*P2*fsI1N2
    return fs

